unit ExpressionClasses;

interface

uses
  Classes, SysUtils;

type
  EExpressionError = class(Exception);

  TExpression = class
  protected
    Text: String;
  public
    function Evaluate(const Strings: TStrings): Boolean; virtual; abstract;
    procedure Print(Indent: Integer; Lines: TStrings); virtual; abstract;
  end;

  TStringExpr = class(TExpression)
  public
    function Evaluate(const Strings: TStrings): Boolean; override;
    procedure Print(Indent: Integer; Lines: TStrings); override;
    constructor Create(const S: String);
  end;

  TBinOpExpr = class(TExpression)
  protected
    LeftExpr, RightExpr: TExpression;
  public
    function Evaluate(const Strings: TStrings): Boolean; override;
    procedure Print(Indent: Integer; Lines: TStrings); override;
    constructor Create(Left: TExpression; const BinOp: String; Right: TExpression);
    destructor Destroy; override;
  end;

  TUnOpExpr = class(TExpression)
  protected
    SubExpr: TExpression;
  public
    function Evaluate(const Strings: TStrings): Boolean; override;
    procedure Print(Indent: Integer; Lines: TStrings); override;
    constructor Create(const UnOp: String; Right: TExpression);
    destructor Destroy; override;
  end;

implementation

function TStringExpr.Evaluate(const Strings: TStrings): Boolean;
begin
  Result := Strings.IndexOf(Text) <> -1
end;

procedure TStringExpr.Print(Indent: Integer; Lines: TStrings);
begin
  Lines.Add(StringOfChar(' ', Indent) + '"' + Text + '"');
end;

constructor TStringExpr.Create(const S: String);
begin
  Text := S;
end;

function TBinOpExpr.Evaluate(const Strings: TStrings): Boolean;
begin
  if SameText(Text, 'and') then
    Result := LeftExpr.Evaluate(Strings) and RightExpr.Evaluate(Strings)
  else
    Result := LeftExpr.Evaluate(Strings) or RightExpr.Evaluate(Strings);
end;

procedure TBinOpExpr.Print(Indent: Integer; Lines: TStrings);
begin
  Lines.Add(StringOfChar(' ', Indent) + Text);
  LeftExpr.Print(Indent+2, Lines);
  RightExpr.Print(Indent+2, Lines);
end;

constructor TBinOpExpr.Create(Left: TExpression; const BinOp: String; Right: TExpression);
begin
  LeftExpr := Left;
  Text := BinOp;
  RightExpr := Right;
end;

destructor TBinOpExpr.Destroy;
begin
  LeftExpr.Free;
  RightExpr.Free;
  inherited;
end;

function TUnOpExpr.Evaluate(const Strings: TStrings): Boolean;
begin
  if AnsiSameText(Text, 'not') then
    Result := not SubExpr.Evaluate(Strings)
  else
    Raise EExpressionError.Create('"Not" is the only allowed unary operator');
end;

procedure TUnOpExpr.Print(Indent: Integer; Lines: TStrings);
begin
  Lines.Add(StringOfChar(' ', Indent) + Text);
  SubExpr.Print(Indent+2, Lines);
end;

constructor TUnOpExpr.Create(const UnOp: String; Right: TExpression);
begin
  Text := UnOp;
  SubExpr := Right;
end;

destructor TUnOpExpr.Destroy;
begin
  SubExpr.Free;
  inherited;
end;

end.
