unit PartialDelphiTokenizationClass;

interface

uses
  TokenizationClass;

type
  TPartialDelphiTokenizationState = (pdtsParsingNothing, pdtsParsingIdent, pdtsParsingDelim,
    pdtsParsingNumber);

  TPartialDelphiTokenization = class(TTokenization)
  protected
    State: TPartialDelphiTokenizationState;
    function BeginsDelim(const S: String): Boolean;
    function EqualsDelim(const S: String): Boolean;
    function IsDigit(C: Char): Boolean;
    function IsIdentCharacter(C: Char): Boolean;
    function IsIdentLeadingCharacter(C: Char): Boolean;
  public
    procedure Initialize; override;
    function ProcessCharacter(C: Char; HandleToken: TTokenHandler; var Reverse: Integer): Boolean; override;
  end;

implementation

uses SysUtils;

const
  // note: no colon, so that ":=" is OK, but ":" is not
  Delims: Array[0..20] of String = ('<>', '<=', '>=', ':=', '.', ',', ';', '(', ')', '{', '}', '[',
    ']', '!', '=', '<', '>', '+', '-', '*', '/');

function TPartialDelphiTokenization.BeginsDelim(const S: String): Boolean;
var
  i: Integer;
begin
  Result := False;
  for i := Low(Delims) to High(Delims) do
    if (Pos(S, Delims[i]) = 1) and (S <> Delims[i]) then
    begin
      Result := True;
      Break;
    end;
end;

function TPartialDelphiTokenization.EqualsDelim(const S: String): Boolean;
var
  i: Integer;
begin
  Result := False;
  for i := Low(Delims) to High(Delims) do
    if S = Delims[i] then
    begin
      Result := True;
      Break;
    end;
end;

function TPartialDelphiTokenization.IsDigit(C: Char): Boolean;
begin
  Result := C in ['0'..'9'];
end;

function TPartialDelphiTokenization.IsIdentCharacter(C: Char): Boolean;
begin
  Result := C in ['_', 'A'..'Z', 'a'..'z', '0'..'9'];
end;

function TPartialDelphiTokenization.IsIdentLeadingCharacter(C: Char): Boolean;
begin
  Result := C in ['_', 'A'..'Z', 'a'..'z'];
end;

procedure TPartialDelphiTokenization.Initialize;
begin
  inherited;
  State := pdtsParsingNothing;
end;

function TPartialDelphiTokenization.ProcessCharacter(C: Char; HandleToken: TTokenHandler;
  var Reverse: Integer): Boolean;
begin
  Result := True;
  case State of
    pdtsParsingNothing:
    begin
      if BeginsDelim(C) then
      begin
        AddToToken(C);
        State := pdtsParsingDelim;
      end
      else if EqualsDelim(C) then
      begin
        DoToken(HandleToken, C);
        AddToToken(C);
        DoToken(HandleToken, C);
      end
      else if IsIdentLeadingCharacter(C) then
      begin
        AddToToken(C);
        State := pdtsParsingIdent;
      end
      else if IsDigit(C) then
      begin
        AddToToken(C);
        State := pdtsParsingNumber;
      end
      else if C in [#0, #10, #13, ' '] then
        DoToken(HandleToken, C)
      else
      begin
        Result := False;
        Exit;
      end;
    end;
    pdtsParsingIdent:
    begin
      if IsIdentCharacter(C) then
        AddToToken(C)
      else if EqualsDelim(C) then
      begin
        DoToken(HandleToken, C);
        AddToToken(C);
        DoToken(HandleToken, C);
        State := pdtsParsingNothing;
      end
      else if C in [#0, #10, #13, ' '] then
      begin
        DoToken(HandleToken, C);
        State := pdtsParsingNothing;
      end
      else if BeginsDelim(C) then
      begin
        DoToken(HandleToken, C);
        AddToToken(C);
        State := pdtsParsingDelim;
      end
      else
      begin
        Result := False;
        Exit;
      end;
    end;
    pdtsParsingDelim:
    begin
      if EqualsDelim(CurrentToken + C) then
      begin
        AddToToken(C);
        DoToken(HandleToken, C);
        State := pdtsParsingNothing;
      end
      else if BeginsDelim(CurrentToken + C) then
        AddToToken(C)
      else if EqualsDelim(CurrentToken) then
      begin
        DoToken(HandleToken, C);
        Reverse := 1;
        State := pdtsParsingNothing;
      end
      else
      begin
        Result := False;
        Exit;
      end;
    end;
    pdtsParsingNumber:
    begin
      if IsDigit(C) then
        AddToToken(C)
      else if C in [#0, #10, #13, ' '] then
      begin
        DoToken(HandleToken, C);
        State := pdtsParsingNothing;
      end
      else if EqualsDelim(C) then
      begin
        DoToken(HandleToken, C);
        AddToToken(C);
        DoToken(HandleToken, C);
        State := pdtsParsingNothing;
      end
      else if BeginsDelim(C) then
      begin
        DoToken(HandleToken, C);
        AddToToken(C);
        State := pdtsParsingDelim;
      end
      else
      begin
        Result := False;
        Exit;
      end;
    end;
  end;
end;

end.
