unit StringEvaluatorClass;

interface

uses
  Classes, ExpressionClasses;

type
  TStringEvaluator = class
  protected
    Tokens: TStrings;
    TokenIndex: Integer;
    ParseTree: TExpression;
    function NextTokenEquals(const Expected: String): Boolean;
    function ThereAreMoreTokens: Boolean;
    function ParseNext: String;

    function NextToken: String;
    function NextIsUnOp: Boolean;
    function NextIsBinOp: Boolean;
    function NextIsLeftPar: Boolean;
    function NextIsRightPar: Boolean;
    function ParseBinaryExpression: TExpression;
    function ParseUnaryExpression: TExpression;
    function ParseParentheses: TExpression;
    function ParseString: TExpression;
    procedure HandleToken(Sender: TObject; const Token: String; Line: Integer);
  public
    procedure CreateBooleanTree(const AText: String);
    function Evaluate(Strings: TStrings): Boolean;
    procedure PrintTree(Lines: TStrings);
    constructor Create;
    destructor Destroy; override;
  end;

implementation

uses
  SysUtils, TokenizerClass, StringEvaluatorTokenizationClass;

function TStringEvaluator.NextToken: String;
begin
  if TokenIndex < Tokens.Count then
    Result := Tokens[TokenIndex]
  else
    Result := '';
end;

function TStringEvaluator.NextTokenEquals(const Expected: String): Boolean;
begin
  if TokenIndex < Tokens.Count then
    Result := SameText(Tokens[TokenIndex], Expected)
  else
    Result := False;
end;

function TStringEvaluator.ThereAreMoreTokens: Boolean;
begin
  Result := TokenIndex < Tokens.Count;
end;

function TStringEvaluator.ParseNext: String;
begin
  if TokenIndex < Tokens.Count then
  begin
    Result := Tokens[TokenIndex];
    Inc(TokenIndex);
  end{if}
  else
    EExpressionError.Create('Unexpected end of input');
end;

function TStringEvaluator.NextIsUnOp: Boolean;
begin
  Result := NextTokenEquals('NOT');
end;

function TStringEvaluator.NextIsBinOp: Boolean;
begin
  Result := NextTokenEquals('AND') or NextTokenEquals('OR');
end;

function TStringEvaluator.NextIsLeftPar: Boolean;
begin
  Result := NextTokenEquals('(');
end;

function TStringEvaluator.NextIsRightPar: Boolean;
begin
  Result := NextTokenEquals(')');
end;

function TStringEvaluator.ParseBinaryExpression: TExpression;
var
  Operator: String;
  Left, Right: TExpression;
begin
  Left := ParseUnaryExpression;
  if ThereAreMoreTokens then
  begin
    while NextIsBinOp do
    begin
      Operator := ParseNext;
      Right := ParseBinaryExpression;
      Left := TBinOpExpr.Create(Left, Operator, Right);
    end;
  end;
  Result := Left;
end;

function TStringEvaluator.ParseUnaryExpression: TExpression;
var
  Operator: String;
  SubExpression: TExpression;
begin
  if ThereAreMoreTokens then
  begin
    if NextTokenEquals('NOT') then
      Operator := ParseNext
    else
      Operator := '';

    SubExpression := ParseParentheses;
    if Operator = '' then
      Result := SubExpression
    else
      Result := TUnOpExpr.Create(Operator, SubExpression);
  end
  else
    Raise EExpressionError.Create('Unexpected end of input');
end;

function TStringEvaluator.ParseParentheses: TExpression;
begin
  if NextTokenEquals('(') then
  begin
    ParseNext;
    Result := ParseBinaryExpression;
    if NextTokenEquals(')') then
      ParseNext
    else
      Raise EExpressionError.Create(Format('Invalid token "%s". Expected ")".', [NextToken]));
  end{if}
  else
    Result := ParseString;
end;

function TStringEvaluator.ParseString: TExpression;
begin
  Result := TStringExpr.Create(ParseNext);
end;

procedure TStringEvaluator.HandleToken(Sender: TObject; const Token: String; Line: Integer);
begin
  Tokens.Add(Token);
end;

procedure TStringEvaluator.CreateBooleanTree(const AText: String);
var
  Tokenizer: TTokenizer;
  Tokenization: TStringEvaluatorTokenization;
begin
  Tokenizer := TTokenizer.Create;
  Tokenization := TStringEvaluatorTokenization.Create;
  try
    Tokens.Clear;
    TokenIndex := 0;
    Tokenizer.ProcessString(AText, Tokenization, HandleToken);
    ParseTree.Free;
    ParseTree := ParseBinaryExpression;
    if ThereAreMoreTokens then
      EExpressionError.Create(Format('Unexpected token "%s".', [NextToken]));
  finally
    Tokenization.Free;
    Tokenizer.Free;
  end;
end;

function TStringEvaluator.Evaluate(Strings: TStrings): Boolean;
begin
  if Assigned(ParseTree) then
    Result := ParseTree.Evaluate(Strings)
  else
    Result := False;
end;

procedure TStringEvaluator.PrintTree(Lines: TStrings);
begin
  Lines.Clear;
  ParseTree.Print(0, Lines);
end;

constructor TStringEvaluator.Create;
begin
  ParseTree := nil;
  Tokens := TStringList.Create;
end;

destructor TStringEvaluator.Destroy;
begin
  ParseTree.Free;
  Tokens.Free;
  inherited;
end;

end.
